{
  Autor: Michael Springwald

  Datum: Montag, 11.Januar.2016

}

unit uplTextEngine;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, LMessages, LCLType, LCLIntf, LCLProc, Messages, Graphics;

type

  { TPLTextEngine }
  TPLTextEngine = class(TCustomControl)
  private
    fTextLineHeight: integer;

  protected
    BarHorz: TScrollInfo;
    BarVert: TScrollInfo;

    procedure WMEraseBkgnd(var Message: TLMEraseBkgnd); message LM_ERASEBKGND;
    procedure EraseBackground(DC: HDC); override;

    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;

  public
    StringList:TStringList;
    ScrollLeft, ScrollTop:Integer;
    ShowBarVert:Boolean;

    function GetLineHeight:Integer;

    constructor Create(AOwner: TComponent);
    destructor Destroy; override;

    procedure Paint; override;
    procedure BoundsChanged; override;
  published
    property TextLineHeight:integer read fTextLineHeight;
  end; // TPLTextEngine

implementation

procedure TPLTextEngine.WMEraseBkgnd(var Message: TLMEraseBkgnd);
begin
  Message.Result := -1;
end; // TPLTextEngine.WMEraseBkgnd

procedure TPLTextEngine.EraseBackground(DC: HDC);
begin
//  inherited EraseBackground(DC);
end; // TPLTextEngine.EraseBackground

procedure TPLTextEngine.WMVScroll(var Message: TWMVScroll);
var
  NewPaint:Boolean;
const
  PIXEL_TO_SCROLL = 20;
begin
  NewPaint := False;
  case message.ScrollCode of
    SB_LINEUP: begin
      if ScrollTop - PIXEL_TO_SCROLL >= 0 then begin
        ScrollTop := ScrollTop - PIXEL_TO_SCROLL;
      end
      else begin
        NewPaint := not (ScrollTop = 0);
        ScrollTop := 0;
      end;
    end; // SB_LINEUP

    SB_LINEDOWN: begin
      if ScrollTop + PIXEL_TO_SCROLL < TextLineHeight - ClientHeight then begin
        ScrollTop := ScrollTop + PIXEL_TO_SCROLL;
        NewPaint := True;
      end
      else begin
        NewPaint := not (ScrollTop = TextLineHeight - ClientHeight);
        ScrollTop := Round(TextLineHeight) - ClientHeight;
      end;
    end; // SB_LINEDOWN

    SB_PAGEDOWN: begin
      if ScrollTop + ClientHeight < TextLineHeight - ClientHeight then begin
        ScrollTop := ScrollTop + ClientHeight;
        NewPaint:=True;
      end
      else begin
        NewPaint := not (ScrollTop = TextLineHeight - ClientHeight);
        ScrollTop := Round(TextLineHeight) - ClientHeight;
      end;
    end; // SB_PAGEDOWN

    SB_PAGEUP: begin
      if ScrollTop - ClientHeight >= 0 then begin
        ScrollTop := ScrollTop - ClientHeight;
        NewPaint := True;
      end
      else begin
        NewPaint := not (ScrollTop = 0);
        ScrollTop := 0;
      end;
    end; // SB_PAGEUP

    SB_THUMBPOSITION,SB_THUMBTRACK : begin
      NewPaint := True;
      ScrollTop :=Message.Pos;
    end; // SB_THUMBPOSITION,SB_THUMBTRACK
  end; // case

  if NewPaint then begin
    BarVert.nPos:=ScrollTop;
    SetScrollInfo(Handle, SB_VERT, BarVert, BarVert.nMax > ClientHeight);
    SetFocus;
    Invalidate;
  end;
end; // TPLTextEngine.WMVScroll

procedure TPLTextEngine.WMHScroll(var Message: TWMHScroll);
begin

end; // TPLTextEngine.WMHScroll

function TPLTextEngine.GetLineHeight: Integer;
var
  i:integer;
begin
  result:=0;
  for i:=0 to StringList.Count-1 do begin
    result:=result+Canvas.TextHeight(StringList[i]);
  end;
  result:=result+10;
end; // TPLTextEngine.GetLineHeight

{ TPLTextEngine }
constructor TPLTextEngine.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  ShowBarVert:=False;

  BarVert.cbSize := SizeOf(BarVert);
  BarVert.fMask := SIF_ALL;  //FMASK;
  BarVert.nMin := 0;

  StringList:=TStringList.Create;
  ScrollLeft:=0;
  ScrollTop:=0;
  fTextLineHeight:=0;
end; // TPLTextEngine.Create

destructor TPLTextEngine.Destroy;
begin
  StringList.Free;
  inherited Destroy;
end; // TPLTextEngine.Destroy

procedure TPLTextEngine.Paint;
var
  px,py,i:Integer;
begin
  inherited Paint;
  if fTextLineHeight = 0 then begin
    fTextLineHeight:=GetLineHeight;
  end;

  px:=5; py:=5;
  Canvas.Brush.Color:=clWindow;
  Canvas.FillRect(0,0,ClientWidth,ClientHeight);
  for i:=0 to StringList.Count-1 do begin
    Canvas.TextOut(px,py-ScrollTop,StringList[i]);
    py:=py+Canvas.TextHeight(StringList[i]);
  end;
end; // TPLTextEngine.Paint

procedure TPLTextEngine.BoundsChanged;
begin
  inherited BoundsChanged;
  BarVert.nPage:=0;
  ShowScrollBar(Handle, SB_HORZ, False);
  ShowScrollBar(Handle, SB_VERT, False);
//  SetScrollInfo(fCustomControl.Handle, SB_HORZ, BarHorz, ShowBarHorz);
  SetScrollInfo(Handle, SB_VERT, BarVert, ShowBarVert);

 {
  ScrollBox.SetBarHorz(Round(MainBox.BodyBox.Width));
  if ScrollBox.BarHorz.nMax > MainBox.ClientWidth then begin
    ScrollBox.fBarHorz.nPage:=MainBox.ClientWidth;
    ScrollBox.ShowBarHorz := ScrollBox.BarHorz.nMax > MainBox.ClientWidth;
    if (ScrollBox.ShowBarHorz) {<> (ScrollBox.BarHorz.nMax > MainBox.ClientWidth)} then
      ShowScrollBar(ScrollBox.fCustomControl.Handle, SB_HORZ, ScrollBox.BarHorz.nMax > MainBox.ClientWidth);
    SetScrollInfo(ScrollBox.fCustomControl.Handle, SB_HORZ, ScrollBox.BarHorz, ScrollBox.ShowBarHorz);
  end;
}

  fTextLineHeight:=GetLineHeight;
  BarVert.nMax := TextLineHeight;
  BarVert.nPos := ScrollTop;

  if BarVert.nMax > ClientHeight then begin
    BarVert.nPage:=ClientHeight;
    ShowBarVert := BarVert.nMax > ClientHeight;

    if (ShowBarVert) then
      ShowScrollBar(Handle, SB_VERT, True);
    SetScrollInfo(Handle, SB_VERT, BarVert, ShowBarVert);
  end;
end;

end.

